from anesthetic.weighted_pandas import WeightedDataFrame, WeightedSeries
from pandas import DataFrame, MultiIndex
import pandas.testing
from anesthetic.utils import channel_capacity
import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
import matplotlib.pyplot as plt
from matplotlib.colors import to_rgba
from pandas.plotting import scatter_matrix, bootstrap_plot
from pandas.plotting._matplotlib.misc import (
    scatter_matrix as orig_scatter_matrix
)
import pandas._testing as tm


@pytest.fixture(autouse=True)
def close_figures_on_teardown():
    tm.close()


@pytest.fixture
def series():
    np.random.seed(0)
    N = 100000
    data = np.random.rand(N)

    series = WeightedSeries(data)
    assert_array_equal(series.get_weights(), 1)
    assert not series.isweighted()
    assert_array_equal(series, data)

    series = WeightedSeries(data, weights=None)
    assert_array_equal(series.get_weights(), 1)
    assert not series.isweighted()
    assert_array_equal(series, data)

    weights = np.random.rand(N)
    series = WeightedSeries(data, weights=weights)
    assert series.isweighted()
    assert_array_equal(series, data)

    assert series.get_weights().shape == (N,)
    assert series.shape == (N,)
    assert isinstance(series.get_weights(), np.ndarray)
    assert_array_equal(series, data)
    assert_array_equal(series.get_weights(), weights)
    assert isinstance(series.to_frame(), WeightedDataFrame)
    assert_array_equal(series.to_frame().get_weights(), weights)
    assert_array_equal(series.index.get_level_values('weights'), weights)

    return series


@pytest.fixture
def frame():
    np.random.seed(0)
    N = 100000
    cols = ['A', 'B', 'C', 'D', 'E', 'F']
    m = len(cols)
    data = np.random.rand(N, m)

    frame = WeightedDataFrame(data, columns=cols)
    assert_array_equal(frame.get_weights(), 1)
    assert not frame.isweighted(0) and not frame.isweighted(1)
    assert_array_equal(frame, data)

    frame = WeightedDataFrame(data, weights=None, columns=cols)
    assert_array_equal(frame.get_weights(), 1)
    assert not frame.isweighted(0) and not frame.isweighted(1)
    assert_array_equal(frame, data)

    weights = np.random.rand(N)
    frame = WeightedDataFrame(data, weights=weights, columns=cols)
    assert frame.isweighted(0) and not frame.isweighted(1)
    assert frame.get_weights().shape == (N,)
    assert frame.shape == (N, m)
    assert isinstance(frame.get_weights(), np.ndarray)
    assert_array_equal(frame, data)
    assert_array_equal(frame.get_weights(), weights)
    assert_array_equal(frame.columns, cols)
    assert_array_equal(frame.index.get_level_values('weights'), weights)
    return frame


def test_WeightedDataFrame_key(frame):
    for key1 in frame.columns:
        assert_array_equal(frame.get_weights(), frame[key1].get_weights())
        for key2 in frame.columns:
            assert_array_equal(frame[key1].get_weights(),
                               frame[key2].get_weights())


def test_WeightedDataFrame_slice(frame):
    assert isinstance(frame['A'], WeightedSeries)
    assert isinstance(frame.iloc[0], WeightedSeries)
    assert not frame.iloc[0].isweighted()
    assert isinstance(frame[:10], WeightedDataFrame)
    assert frame[:10].isweighted()
    assert frame[:10].shape == (10, 6)
    assert frame[:10].get_weights().shape == (10,)
    assert frame[:10]._rand(0).shape == (10,)
    assert frame[:10]._rand(1).shape == (6,)


def test_WeightedDataFrame_mean(frame):
    mean = frame.mean()
    assert isinstance(mean, WeightedSeries)
    assert not mean.isweighted()
    assert_array_equal(mean.index, frame.columns)
    assert_allclose(mean, 0.5, atol=1e-2)

    mean = frame.mean(axis=1)
    assert isinstance(mean, WeightedSeries)
    assert_array_equal(mean, frame.T.mean())
    assert isinstance(frame.T.mean(), WeightedSeries)
    assert mean.isweighted()
    assert frame.T.mean().isweighted()
    assert_array_equal(mean.index, frame.index)
    assert_allclose(mean.mean(), 0.5, atol=1e-2)


def test_WeightedDataFrame_std(frame):
    std = frame.std()
    assert isinstance(std, WeightedSeries)
    assert not std.isweighted()
    assert_array_equal(std.index, frame.columns)
    assert_allclose(std, (1./12)**0.5, atol=1e-2)

    std = frame.std(axis=1)
    assert isinstance(std, WeightedSeries)
    assert_array_equal(std, frame.T.std())
    assert isinstance(frame.T.std(), WeightedSeries)
    assert std.isweighted()
    assert frame.T.std().isweighted()
    assert_array_equal(std.index, frame.index)
    assert_allclose(std.mean(), (1./12)**0.5, atol=1e-1)


def test_WeightedDataFrame_cov(frame):
    cov = frame.cov()
    assert isinstance(cov, WeightedDataFrame)
    assert not cov.isweighted(0) and not cov.isweighted(1)
    assert_array_equal(cov.index, frame.columns)
    assert_array_equal(cov.columns, frame.columns)
    assert_allclose(cov, (1./12)*np.identity(6), atol=1e-2)

    cov = frame[:5].T.cov()
    assert isinstance(cov, WeightedDataFrame)
    assert cov.isweighted(0) and cov.isweighted(1)
    assert_array_equal(cov.index, frame[:5].index)
    assert_array_equal(cov.columns, frame[:5].index)

    with pytest.raises(NotImplementedError):
        # kwargs not passed when weighted
        frame.cov(ddof=1)


def test_WeightedDataFrame_corr(frame):
    corr = frame.corr()
    assert isinstance(corr, WeightedDataFrame)
    assert not corr.isweighted(0) and not corr.isweighted(1)
    assert_array_equal(corr.index, frame.columns)
    assert_array_equal(corr.columns, frame.columns)
    assert_allclose(corr, np.identity(6), atol=1.1e-2)

    corr = frame[:5].T.corr()
    assert isinstance(corr, WeightedDataFrame)
    assert corr.isweighted(0) and corr.isweighted(1)
    assert_array_equal(corr.index, frame[:5].index)
    assert_array_equal(corr.columns, frame[:5].index)


def test_WeightedDataFrame_corrwith(frame):
    correl = frame.corrwith(frame.A)
    assert isinstance(correl, WeightedSeries)
    assert not correl.isweighted()
    assert_array_equal(correl.index, frame.columns)
    assert_allclose(correl, frame.corr()['A'], atol=1e-2)

    correl = frame.corrwith(frame[['A', 'B']])
    assert isinstance(correl, WeightedSeries)
    assert not correl.isweighted()
    assert_array_equal(correl.index, frame.columns)
    assert_allclose(correl['A'], 1, atol=1e-2)
    assert_allclose(correl['B'], 1, atol=1e-2)
    assert np.isnan(correl['C'])

    unweighted = DataFrame(frame).droplevel('weights')

    with pytest.raises(ValueError):
        frame.corrwith(unweighted.A)

    with pytest.raises(ValueError):
        frame.corrwith(unweighted[['A', 'B']])

    with pytest.raises(ValueError):
        unweighted.corrwith(frame[['A', 'B']])

    correl_1 = unweighted[:5].corrwith(unweighted[:4], axis=1)
    correl_2 = frame[:5].corrwith(frame[:4], axis=1)
    assert_array_equal(correl_1.values, correl_2.values)
    assert correl_2.isweighted()

    correl_3 = frame[:5].T.corrwith(frame[:4].T)
    assert_array_equal(correl_2, correl_3)
    assert_array_equal(correl_2.index, correl_3.index)

    correl_4 = frame.T.corrwith(frame.T, axis=1)
    correl_5 = unweighted.T.corrwith(unweighted.T, axis=1)
    assert_allclose(correl_4, correl_5)

    frame.set_weights(None, inplace=True)
    assert_array_equal(frame.corrwith(frame), unweighted.corrwith(unweighted))


def test_WeightedDataFrame_median(frame):
    median = frame.median()
    assert isinstance(median, WeightedSeries)
    assert not median.isweighted()
    assert_array_equal(median.index, frame.columns)
    assert_allclose(median, 0.5, atol=1e-2)

    median = frame.median(axis=1)
    assert isinstance(median, WeightedSeries)
    assert_array_equal(median, frame.T.median())
    assert isinstance(frame.T.median(), WeightedSeries)
    assert median.isweighted()
    assert frame.T.median().isweighted()
    assert_array_equal(median.index, frame.index)
    assert_allclose(median.mean(), 0.5, atol=1e-2)


def test_WeightedDataFrame_sem(frame):
    sem = frame.sem()
    assert isinstance(sem, WeightedSeries)
    assert not sem.isweighted()
    assert_array_equal(sem.index, frame.columns)
    assert_allclose(sem, (1./12)**0.5/np.sqrt(frame.neff()), atol=1e-2)

    sem = frame.sem(axis=1)
    assert isinstance(sem, WeightedSeries)
    assert_array_equal(sem, frame.T.sem())
    assert isinstance(frame.T.sem(), WeightedSeries)
    assert sem.isweighted()
    assert frame.T.sem().isweighted()
    assert_array_equal(sem.index, frame.index)


def test_WeightedDataFrame_kurtosis(frame):
    kurtosis = frame.kurtosis()
    assert isinstance(kurtosis, WeightedSeries)
    assert not kurtosis.isweighted()
    assert_array_equal(kurtosis.index, frame.columns)
    assert_allclose(kurtosis, 9./5, atol=1e-2)
    assert_array_equal(frame.kurtosis(), frame.kurt())

    kurtosis = frame.kurtosis(axis=1)
    assert isinstance(kurtosis, WeightedSeries)
    assert_array_equal(kurtosis, frame.T.kurtosis())
    assert isinstance(frame.T.kurtosis(), WeightedSeries)
    assert kurtosis.isweighted()
    assert frame.T.kurtosis().isweighted()
    assert_array_equal(kurtosis.index, frame.index)
    assert_array_equal(frame.kurtosis(axis=1), frame.kurt(axis=1))


def test_WeightedDataFrame_skew(frame):
    skew = frame.skew()
    assert isinstance(skew, WeightedSeries)
    assert not skew.isweighted()
    assert_array_equal(skew.index, frame.columns)
    assert_allclose(skew, 0., atol=2e-2)

    skew = frame.skew(axis=1)
    assert isinstance(skew, WeightedSeries)
    assert_array_equal(skew, frame.T.skew())
    assert isinstance(frame.T.skew(), WeightedSeries)
    assert skew.isweighted()
    assert frame.T.skew().isweighted()
    assert_array_equal(skew.index, frame.index)


def test_WeightedDataFrame_mad(frame):
    mad = frame.mad()
    assert isinstance(mad, WeightedSeries)
    assert not mad.isweighted()
    assert_array_equal(mad.index, frame.columns)
    assert_allclose(mad, 0.25, atol=1e-2)

    mad = frame.mad(axis=1)
    assert isinstance(mad, WeightedSeries)
    assert_array_equal(mad, frame.T.mad())
    assert isinstance(frame.T.mad(), WeightedSeries)
    assert mad.isweighted()
    assert frame.T.mad().isweighted()
    assert_array_equal(mad.index, frame.index)


def test_WeightedDataFrame_quantile(frame):
    quantile = frame.quantile()
    assert isinstance(quantile, WeightedSeries)
    assert not quantile.isweighted()
    assert_array_equal(quantile.index, frame.columns)
    assert_allclose(quantile, 0.5, atol=1e-2)

    quantile = frame.quantile(axis=1)
    assert isinstance(quantile, WeightedSeries)
    assert_array_equal(quantile, frame.T.quantile())
    assert quantile.isweighted()
    assert frame.T.quantile().isweighted()
    assert_array_equal(quantile.index, frame.index)
    assert_allclose(quantile.mean(), 0.5, atol=1e-2)

    qs = np.linspace(0, 1, 10)
    for q in qs:
        quantile = frame.quantile(q)
        assert isinstance(quantile, WeightedSeries)
        assert not quantile.isweighted()
        assert_array_equal(quantile.index, frame.columns)
        assert_allclose(quantile, q, atol=1e-2)

        quantile = frame.quantile(q, axis=1)
        assert isinstance(quantile, WeightedSeries)

        quantile = frame.quantile(q, axis=1)
        assert isinstance(quantile, WeightedSeries)
        assert_array_equal(quantile, frame.T.quantile(q))
        assert quantile.isweighted()
        assert frame.T.quantile(q).isweighted()
        assert_array_equal(quantile.index, frame.index)

    quantile = frame.quantile(qs)
    assert_allclose(quantile, np.transpose([qs]*6), atol=1e-2)
    assert_array_equal(quantile.index, qs)
    assert_array_equal(quantile.columns, frame.columns)
    quantile = frame.quantile(qs, axis=1)
    assert isinstance(quantile, WeightedDataFrame)
    assert_array_equal(quantile.index, qs)
    assert_array_equal(quantile.columns, frame.index)

    with pytest.raises(NotImplementedError):
        frame.quantile(numeric_only=False)
    with pytest.raises(NotImplementedError):
        frame.quantile(method='single')


def test_WeightedDataFrame_sample(frame):
    sample = frame.sample()
    assert isinstance(sample, WeightedDataFrame)
    assert_array_equal(sample.columns, frame.columns)
    assert sample.index.isin(frame.index)
    assert sample.isin(frame).all().all()

    samples = frame.sample(5)
    assert isinstance(samples, WeightedDataFrame)
    assert_array_equal(sample.columns, frame.columns)
    assert samples.index.isin(frame.index).all()
    assert samples.isin(frame).all().all()
    assert len(samples) == 5

    sample = frame.sample(axis=1)
    assert isinstance(sample, WeightedDataFrame)
    assert_array_equal(sample.index, frame.index)
    assert sample.columns.isin(frame.columns).all()
    assert sample.isin(frame).all().all()

    samples = frame.sample(5, axis=1)
    assert isinstance(samples, WeightedDataFrame)
    assert_array_equal(samples.index, frame.index)
    assert samples.columns.isin(frame.columns).all()
    assert samples.isin(frame).all().all()
    assert len(samples.columns) == 5

    frame.T.sample()
    frame.T.sample(5)
    frame.T.sample(axis=1)
    frame.T.sample(5, axis=1)


def test_WeightedDataFrame_neff(frame):
    neff = frame.neff()
    assert isinstance(neff, float)
    assert neff < len(frame)
    assert neff > len(frame) * np.exp(-0.25)

    neff = frame.neff(1)
    assert isinstance(neff, int)
    assert neff == len(frame.T)


def test_WeightedDataFrame_compress(frame):
    assert_allclose(frame.neff(), len(frame.compress()), rtol=1e-2)
    for i in np.logspace(3, 5, 10):
        assert_allclose(i, len(frame.compress(i)), rtol=1e-1)
    unit_weights = frame.compress(0)
    assert len(np.unique(unit_weights.index)) == len(unit_weights)
    assert_array_equal(frame.compress(), frame.compress())
    assert_array_equal(frame.compress(i), frame.compress(i))
    assert_array_equal(frame.compress(-1), frame.compress(-1))

    assert_array_equal(frame.T.compress().T, frame)
    assert_array_equal(frame.T.compress(axis=1).T, frame.compress())


def test_WeightedDataFrame_nan(frame):
    frame['A'][0] = np.nan
    assert ~frame.mean().isna().any()
    assert ~frame.mean(axis=1).isna().any()
    assert_array_equal(frame.mean(skipna=False).isna(), [True] + [False]*5)
    assert_array_equal(frame.mean(axis=1, skipna=False).isna()[0:6],
                       [True, False, False, False, False, False])

    assert ~frame.std().isna().any()
    assert ~frame.std(axis=1).isna().any()
    assert_array_equal(frame.std(skipna=False).isna(), [True] + [False]*5)
    assert_array_equal(frame.std(axis=1, skipna=False).isna()[0:6],
                       [True, False, False, False, False, False])

    frame['B'][2] = np.nan
    assert ~frame.mean().isna().any()
    assert_array_equal(frame.mean(skipna=False).isna(),
                       [True, True] + [False]*4)
    assert_array_equal(frame.mean(axis=1, skipna=False).isna()[0:6],
                       [True, False, True, False, False, False])

    assert ~frame.std().isna().any()
    assert_array_equal(frame.std(skipna=False).isna(),
                       [True, True] + [False]*4)
    assert_array_equal(frame.std(axis=1, skipna=False).isna()[0:6],
                       [True, False, True, False, False, False])

    frame['C'][4] = np.nan
    frame['D'][5] = np.nan
    frame['E'][6] = np.nan
    frame['F'][7] = np.nan
    assert ~frame.mean().isna().any()
    assert frame.mean(skipna=False).isna().all()
    assert_array_equal(frame.mean(axis=1, skipna=False).isna()[0:6],
                       [True, False, True, False, True, True])

    assert ~frame.std().isna().any()
    assert frame.std(skipna=False).isna().all()
    assert_array_equal(frame.std(axis=1, skipna=False).isna()[0:6],
                       [True, False, True, False, True, True])

    assert_allclose(frame.mean(), 0.5, atol=1e-2)
    assert_allclose(frame.std(), (1./12)**0.5, atol=1e-2)
    assert_allclose(frame.cov(), (1./12)*np.identity(6), atol=1e-2)

    assert isinstance(frame.mean(), WeightedSeries)
    assert not frame.mean().isweighted()
    assert isinstance(frame.mean(axis=1), WeightedSeries)
    assert frame.mean(axis=1).isweighted()

    assert frame[:0].mean().isna().all()
    assert frame[:0].std().isna().all()
    assert frame[:0].median().isna().all()
    assert frame[:0].var().isna().all()
    assert frame[:0].cov().isna().all().all()
    assert frame[:0].corr().isna().all().all()
    assert frame[:0].kurt().isna().all()
    assert frame[:0].skew().isna().all()
    assert frame[:0].mad().isna().all()
    assert frame[:0].sem().isna().all()
    assert frame[:0].quantile().isna().all()


def test_WeightedSeries_mean(series):
    series[0] = np.nan
    series.var(skipna=False)
    mean = series.mean()
    assert isinstance(mean, float)
    assert_allclose(mean, 0.5, atol=1e-2)


def test_WeightedSeries_std(series):
    std = series.std()
    assert isinstance(std, float)
    assert_allclose(std, (1./12)**0.5, atol=1e-2)

    series[0] = np.nan
    assert ~np.isnan(series.std())
    assert np.isnan(series.std(skipna=False))


def test_WeightedSeries_cov(frame):
    assert_allclose(frame.A.cov(frame.A), 1./12, atol=1e-2)
    assert_allclose(frame.A.cov(frame.B), 0, atol=1e-2)

    frame['A'][0] = np.nan
    assert_allclose(frame.A.cov(frame.A), 1./12, atol=1e-2)
    assert_allclose(frame.A.cov(frame.B), 0, atol=1e-2)


def test_WeightedSeries_corr(frame):
    assert_allclose(frame.A.corr(frame.A), 1., atol=1e-2)
    assert_allclose(frame.A.corr(frame.B), 0, atol=1e-2)
    D = frame.A + frame.B
    assert_allclose(frame.A.corr(D), 1/np.sqrt(2), atol=1e-2)

    unweighted = DataFrame(frame).droplevel('weights')

    with pytest.raises(ValueError):
        frame.A.corr(unweighted.B)

    with pytest.raises(ValueError):
        unweighted.A.corr(frame.B)


def test_WeightedSeries_median(series):
    median = series.median()
    assert isinstance(median, float)
    assert_allclose(median, 0.5, atol=1e-2)


def test_WeightedSeries_sem(series):
    sem = series.sem()
    assert isinstance(sem, float)
    assert_allclose(sem, (1./12)**0.5/np.sqrt(series.neff()), atol=1e-2)


def test_WeightedSeries_kurtosis(series):
    kurtosis = series.kurtosis()
    assert isinstance(kurtosis, float)
    assert_allclose(kurtosis, 9./5, atol=1e-2)
    assert series.kurtosis() == series.kurt()

    series[0] = np.nan
    assert ~np.isnan(series.kurtosis())
    assert np.isnan(series.kurtosis(skipna=False))


def test_WeightedSeries_skew(series):
    skew = series.skew()
    assert isinstance(skew, float)
    assert_allclose(skew, 0., atol=1e-2)

    series[0] = np.nan
    assert ~np.isnan(series.skew())
    assert np.isnan(series.skew(skipna=False))


def test_WeightedSeries_mad(series):
    mad = series.mad()
    assert isinstance(mad, float)
    assert_allclose(mad, 0.25, atol=1e-2)

    series[0] = np.nan
    assert ~np.isnan(series.mad())
    assert np.isnan(series.mad(skipna=False))


def test_WeightedSeries_quantile(series):

    quantile = series.quantile()
    assert isinstance(quantile, float)
    assert_allclose(quantile, 0.5, atol=1e-2)

    qs = np.linspace(0, 1, 10)
    for q in qs:
        quantile = series.quantile(q)
        assert isinstance(quantile, float)
        assert_allclose(quantile, q, atol=1e-2)

    assert_allclose(series.quantile(qs), qs, atol=1e-2)


def test_WeightedSeries_sample(series):
    sample = series.sample()
    assert isinstance(sample, WeightedSeries)
    samples = series.sample(5)
    assert isinstance(samples, WeightedSeries)
    assert len(samples) == 5


def test_WeightedSeries_neff(series):
    neff = series.neff()
    assert isinstance(neff, float)
    assert neff < len(series)
    assert neff > len(series) * np.exp(-0.25)


def test_WeightedSeries_compress(series):
    assert_allclose(series.neff(), len(series.compress()), rtol=1e-2)
    for i in np.logspace(3, 5, 10):
        assert_allclose(i, len(series.compress(i)), rtol=1e-1)
    unit_weights = series.compress(0)
    assert len(np.unique(unit_weights.index)) == len(unit_weights)


def test_WeightedSeries_nan(series):

    series[0] = np.nan

    assert ~np.isnan(series.mean())
    assert np.isnan(series.mean(skipna=False))
    assert ~np.isnan(series.std())
    assert np.isnan(series.std(skipna=False))
    assert ~np.isnan(series.var())
    assert np.isnan(series.var(skipna=False))

    assert_allclose(series.mean(), 0.5, atol=1e-2)
    assert_allclose(series.var(), 1./12, atol=1e-2)
    assert_allclose(series.std(), (1./12)**0.5, atol=1e-2)

    assert np.isnan(series[:0].mean())
    assert np.isnan(series[:0].std())
    assert np.isnan(series[:0].median())
    assert np.isnan(series[:0].var())
    assert np.isnan(series[:0].cov(series))
    assert np.isnan(series[:0].corr(series))
    assert np.isnan(series[:0].kurt())
    assert np.isnan(series[:0].skew())
    assert np.isnan(series[:0].mad())
    assert np.isnan(series[:0].sem())
    assert np.isnan(series[:0].quantile())


@pytest.fixture
def mcmc_df():
    np.random.seed(0)
    m, s = 0.5, 0.1

    def logL(x):
        return x, -((x-m)**2).sum(axis=-1)/2/s**2

    np.random.seed(0)
    x0, logL0 = logL(np.random.normal(m, s, 4))
    dat = []
    for _ in range(3000):
        dat.append(x0)
        x1, logL1 = logL(np.random.normal(x0, s/2))
        if np.log(np.random.rand()) < logL1 - logL0:
            x0, logL0 = x1, logL1

    return DataFrame(dat, columns=["x", "y", "z", "w"])


@pytest.fixture
def mcmc_wdf(mcmc_df):
    weights = mcmc_df.groupby(mcmc_df.columns.tolist(), sort=False).size()
    return WeightedDataFrame(mcmc_df.drop_duplicates(), weights=weights.values)


def test_WeightedDataFrame_hist(mcmc_df, mcmc_wdf):
    df_axes = mcmc_df.hist().flatten()
    wdf_axes = mcmc_wdf.hist().flatten()

    df_heights = [p.get_height() for ax in df_axes for p in ax.patches]
    wdf_heights = [p.get_height() for ax in wdf_axes for p in ax.patches]
    assert df_heights == wdf_heights

    df_widths = [p.get_width() for ax in df_axes for p in ax.patches]
    wdf_widths = [p.get_width() for ax in wdf_axes for p in ax.patches]
    assert df_widths == wdf_widths

    df_xys = [p.get_xy() for ax in df_axes for p in ax.patches]
    wdf_xys = [p.get_xy() for ax in wdf_axes for p in ax.patches]
    assert df_xys == wdf_xys

    df_axes = mcmc_df.plot.hist(subplots=True)
    wdf_axes = mcmc_wdf.plot.hist(subplots=True)

    df_heights = [p.get_height() for ax in df_axes for p in ax.patches]
    wdf_heights = [p.get_height() for ax in wdf_axes for p in ax.patches]
    assert df_heights == wdf_heights

    df_widths = [p.get_width() for ax in df_axes for p in ax.patches]
    wdf_widths = [p.get_width() for ax in wdf_axes for p in ax.patches]
    assert df_widths == wdf_widths

    df_xys = [p.get_xy() for ax in df_axes for p in ax.patches]
    wdf_xys = [p.get_xy() for ax in wdf_axes for p in ax.patches]
    assert df_xys == wdf_xys


def test_WeightedSeries_hist(mcmc_df, mcmc_wdf):

    fig, axes = plt.subplots(2)
    mcmc_df.x.hist(ax=axes[0])
    mcmc_wdf.x.hist(ax=axes[1])

    df_heights = [p.get_height() for p in axes[0].patches]
    wdf_heights = [p.get_height() for p in axes[1].patches]
    assert df_heights == wdf_heights

    df_widths = [p.get_width() for p in axes[0].patches]
    wdf_widths = [p.get_width() for p in axes[1].patches]
    assert df_widths == wdf_widths

    df_xys = [p.get_xy() for p in axes[0].patches]
    wdf_xys = [p.get_xy() for p in axes[1].patches]
    assert df_xys == wdf_xys

    fig, axes = plt.subplots(2)
    mcmc_df.x.plot.hist(ax=axes[0])
    mcmc_wdf.x.plot.hist(ax=axes[1])

    df_heights = [p.get_height() for p in axes[0].patches]
    wdf_heights = [p.get_height() for p in axes[1].patches]
    assert df_heights == wdf_heights

    df_widths = [p.get_width() for p in axes[0].patches]
    wdf_widths = [p.get_width() for p in axes[1].patches]
    assert df_widths == wdf_widths

    df_xys = [p.get_xy() for p in axes[0].patches]
    wdf_xys = [p.get_xy() for p in axes[1].patches]
    assert df_xys == wdf_xys


def test_KdePlot(mcmc_df, mcmc_wdf):
    bw_method = 0.3
    fig, axes = plt.subplots(2)
    mcmc_df.x.plot.kde(bw_method=bw_method, ax=axes[0])
    mcmc_wdf.x.plot.kde(bw_method=bw_method, ax=axes[1])
    df_line, wdf_line = axes[0].lines[0], axes[1].lines[0]
    assert (df_line.get_xdata() == wdf_line.get_xdata()).all()
    assert_allclose(df_line.get_ydata(),  wdf_line.get_ydata(), atol=1e-4)


def test_scatter_matrix(mcmc_df, mcmc_wdf):
    axes = scatter_matrix(mcmc_df)
    data = axes[0, 1].collections[0].get_offsets().data
    axes = orig_scatter_matrix(mcmc_df)
    orig_data = axes[0, 1].collections[0].get_offsets().data

    assert_allclose(data, orig_data)

    axes = scatter_matrix(mcmc_wdf)
    data = axes[0, 1].collections[0].get_offsets().data
    n = len(data)
    neff = channel_capacity(mcmc_wdf.get_weights())
    assert_allclose(n, neff, atol=np.sqrt(n))

    axes = orig_scatter_matrix(mcmc_wdf)
    orig_data = axes[0, 1].collections[0].get_offsets().data
    n = len(orig_data)
    assert n == len(mcmc_wdf)

    axes = scatter_matrix(mcmc_wdf, ncompress=50)
    n = len(axes[0, 1].collections[0].get_offsets().data)
    assert_allclose(n, 50, atol=np.sqrt(n))


def test_bootstrap_plot(mcmc_df, mcmc_wdf):
    bootstrap_plot(mcmc_wdf.x)
    bootstrap_plot(mcmc_wdf.x, ncompress=500)


def test_BoxPlot(mcmc_df, mcmc_wdf):
    mcmc_df.plot.box()
    mcmc_wdf.plot.box()

    fig, ax = plt.subplots()
    mcmc_df.boxplot(ax=ax)

    fig, ax = plt.subplots()
    mcmc_wdf.boxplot(ax=ax)

    fig, ax = plt.subplots()
    mcmc_df.x.plot.box(ax=ax)

    fig, ax = plt.subplots()
    mcmc_wdf.x.plot.box(ax=ax)

    mcmc_df.plot.box(subplots=True)
    mcmc_wdf.plot.box(subplots=True)

    mcmc_df['split'] = ''
    mcmc_df.loc[:len(mcmc_df)//2, 'split'] = 'A'
    mcmc_df.loc[len(mcmc_df)//2:, 'split'] = 'B'

    mcmc_wdf['split'] = ''
    mcmc_wdf.iloc[:len(mcmc_wdf)//2, -1] = 'A'
    mcmc_wdf.iloc[len(mcmc_wdf)//2:, -1] = 'B'

    mcmc_df.groupby('split').plot.box()
    mcmc_wdf.groupby('split').plot.box()

    for return_type in ['dict', 'both']:
        fig, ax = plt.subplots()
        mcmc_wdf.plot.box(return_type=return_type, ax=ax)
        fig, ax = plt.subplots()
        mcmc_wdf.boxplot(return_type=return_type, ax=ax)

    fig, ax = plt.subplots()
    mcmc_wdf.boxplot(xlabel='xlabel', ax=ax)
    fig, ax = plt.subplots()
    mcmc_wdf.boxplot(ylabel='ylabel', ax=ax)

    fig, ax = plt.subplots()
    mcmc_wdf.boxplot(vert=False, ax=ax)

    fig, ax = plt.subplots()
    mcmc_wdf.boxplot(fontsize=30, ax=ax)


def test_ScatterPlot(mcmc_df, mcmc_wdf):
    mcmc_df.plot.scatter('x', 'y')
    ax = mcmc_wdf.plot.scatter('x', 'y')

    n = len(ax.collections[0].get_offsets().data)
    neff = channel_capacity(mcmc_wdf.get_weights())
    assert_allclose(n, neff, atol=np.sqrt(n))

    ax = mcmc_wdf.plot.scatter('x', 'y', ncompress=50)
    n = len(ax.collections[0].get_offsets().data)
    assert_allclose(n, 50, atol=np.sqrt(50))


def test_HexBinPlot(mcmc_df, mcmc_wdf):
    df_axes = mcmc_df.plot.hexbin('x', 'y', mincnt=1)
    wdf_axes = mcmc_wdf.plot.hexbin('x', 'y')

    df_data = df_axes.collections[0].get_offsets()
    wdf_data = wdf_axes.collections[0].get_offsets()
    assert_allclose(df_data, wdf_data)

    df_colors = df_axes.collections[0].get_facecolors()
    wdf_colors = wdf_axes.collections[0].get_facecolors()
    assert_allclose(df_colors, wdf_colors)

    plt.close("all")


def test_AreaPlot(mcmc_df, mcmc_wdf):
    fig, (ax1, ax2) = plt.subplots(1, 2)
    axes_df = mcmc_df.plot.area(ax=ax1)
    axes_wdf = mcmc_wdf.plot.area(ax=ax2)

    assert_allclose(axes_df.get_xlim(), axes_wdf.get_xlim(), rtol=1e-3)
    assert_allclose(axes_df.get_ylim(), axes_wdf.get_ylim(), rtol=1e-3)


def test_BarPlot(mcmc_df, mcmc_wdf):
    axes_bar = mcmc_wdf[5:10].plot.bar()
    axes_barh = mcmc_wdf[5:10].plot.barh()
    assert_array_equal(axes_bar.get_xticks(), axes_barh.get_yticks())
    assert_array_equal(axes_bar.get_yticks(), axes_barh.get_xticks())


def test_PiePlot(mcmc_df, mcmc_wdf):
    fig, ax = plt.subplots()
    mcmc_wdf[5:10].x.plot.pie(ax=ax)
    mcmc_wdf[5:10].plot.pie(subplots=True)


def test_LinePlot(mcmc_df, mcmc_wdf):
    fig, ax = plt.subplots()
    df_axes = mcmc_df.x.plot.line(ax=ax)
    assert_array_equal(mcmc_df.index, df_axes.lines[0].get_xdata())
    fig, ax = plt.subplots()
    wdf_axes = mcmc_wdf.x.plot.line(ax=ax)
    assert_array_equal(mcmc_wdf.index.droplevel('weights'),
                       wdf_axes.lines[0].get_xdata())
    wdf_axes = mcmc_wdf.plot.line()
    assert len(wdf_axes.lines) == len(mcmc_wdf.columns)


def test_multiindex(mcmc_wdf):
    np.random.seed(0)
    i1 = np.arange(len(mcmc_wdf.index))
    i2 = np.random.randint(0, len(mcmc_wdf.index), len(mcmc_wdf.index))
    i3 = np.random.randint(0, len(mcmc_wdf.index), len(mcmc_wdf.index))
    index = MultiIndex.from_arrays([i1, i2, i3], names=['A', 'B', 'C'])
    weights = mcmc_wdf.get_weights()
    wdf = WeightedDataFrame(mcmc_wdf.values, weights=weights, index=index)

    assert wdf.index.names == ['A', 'B', 'C', 'weights']
    assert_allclose(np.array([*wdf.index]).T, [i1, i2, i3, weights])

    assert wdf.reset_index().index.names == [None, 'weights']

    assert not np.array_equal(wdf.reset_index().columns, wdf.columns)
    assert_array_equal(wdf.reset_index(drop=True).columns, wdf.columns)

    new = wdf.copy()
    assert not np.array_equal(new.index, wdf.reset_index())
    new.reset_index(inplace=True)
    assert_array_equal(new.index, wdf.reset_index().index)

    wdf_ = wdf.reset_index(level='A')
    assert wdf_.index.names == ['B', 'C', 'weights']
    wdf_ = wdf.reset_index(level=['A', 'C'])
    assert wdf_.index.names == ['B', 'weights']

    assert_array_equal(wdf.get_weights(), weights)
    wdf_ = wdf.reorder_levels(['B', 'C', 'weights', 'A'])
    assert_array_equal(wdf_.get_weights(), weights)
    weights_ = np.random.rand(len(weights))
    assert_array_equal(wdf_.set_weights(weights_).get_weights(), weights_)

    wdf_ = wdf.droplevel('weights').set_weights(weights, level=2)
    assert wdf_.index.names == ['A', 'B', 'weights', 'C']
    assert_array_equal(wdf_.get_weights(), weights)

    wdf_ = wdf.set_weights(weights, level=2)
    assert wdf_.index.names == ['A', 'B', 'weights', 'C']
    assert_array_equal(wdf_.get_weights(), weights)


def test_weight_passing(mcmc_wdf):
    weights = mcmc_wdf.get_weights()
    new_wdf = WeightedDataFrame(mcmc_wdf.copy(), weights=None)
    assert (new_wdf.get_weights() == mcmc_wdf.get_weights()).all()

    np.random.shuffle(weights)
    assert (mcmc_wdf.get_weights() != weights).any()

    new_wdf = WeightedDataFrame(mcmc_wdf.copy(), weights=weights)
    assert_array_equal(new_wdf.get_weights(), weights)


def test_set_weights(mcmc_wdf):
    weights_1 = mcmc_wdf.get_weights()
    weights_2 = np.random.rand(len(mcmc_wdf.index))

    assert_array_equal(mcmc_wdf.set_weights(weights_2).get_weights(),
                       weights_2)
    assert_array_equal(mcmc_wdf.get_weights(), weights_1)
    mcmc_wdf.set_weights(weights_2, inplace=True)
    assert_array_equal(mcmc_wdf.get_weights(), weights_2)

    assert mcmc_wdf.isweighted()
    assert not mcmc_wdf.set_weights(None).isweighted()
    assert_array_equal(mcmc_wdf.set_weights(None).get_weights(), 1)

    mcmc_wdf.set_weights(None, inplace=True)
    assert not mcmc_wdf.isweighted()
    assert mcmc_wdf.set_weights(None) is not mcmc_wdf

    mcmc_id = id(mcmc_wdf)
    mcmc_wdf.set_weights(None, inplace=True)
    assert id(mcmc_wdf) == mcmc_id
    assert not mcmc_wdf.isweighted()


def test_drop_weights(mcmc_wdf):
    assert mcmc_wdf.isweighted()
    noweights = mcmc_wdf.drop_weights()
    assert not noweights.isweighted()
    assert_array_equal(noweights, mcmc_wdf)
    pandas.testing.assert_frame_equal(noweights.drop_weights(), noweights)
    assert noweights.drop_weights() is not noweights


def test_blank_axis_labels(mcmc_df, mcmc_wdf):
    for df in mcmc_df, mcmc_wdf:
        assert df.plot.area().get_xlabel() == ""
        assert df.plot.bar().get_xlabel() == ""
        assert df.plot.barh().get_ylabel() == ""


def test_get_index(mcmc_wdf):
    mcmc_wdf.index = mcmc_wdf.index.rename(('foo', 'weights'))
    assert mcmc_wdf.plot.area().get_xlabel() == "foo"
    assert mcmc_wdf.plot.bar().get_xlabel() == "foo"
    assert mcmc_wdf.plot.barh().get_ylabel() == "foo"
    for plot in [mcmc_wdf.plot.area, mcmc_wdf.plot.bar, mcmc_wdf.plot.barh]:
        assert plot(xlabel="my xlabel").get_xlabel() == "my xlabel"

    idx = mcmc_wdf.index.to_frame().to_numpy().T
    mcmc_wdf.index = MultiIndex.from_arrays([idx[0], idx[0], idx[1]],
                                            names=('foo', 'bar', 'weights'))
    assert mcmc_wdf.plot.bar().get_xlabel() == "foo,bar"
    assert mcmc_wdf.plot.barh().get_ylabel() == "foo,bar"
    mcmc_wdf.index = MultiIndex.from_arrays([idx[0], idx[0], idx[1]],
                                            names=(None, None, 'weights'))
    assert mcmc_wdf.plot.bar().get_xlabel() == ""
    assert mcmc_wdf.plot.barh().get_ylabel() == ""


def test_style(mcmc_wdf):
    style_dict = dict(x='c', y='y', z='m', w='k')
    ax = mcmc_wdf.plot.kde(style=style_dict)
    assert all([line.get_color() == to_rgba(c)
                for line, c in zip(ax.get_lines(), ['c', 'y', 'm', 'k'])])

    with pytest.raises(TypeError):
        ax = mcmc_wdf.plot.hist_2d('x', 'y', style='c')
    with pytest.raises(TypeError):
        ax = mcmc_wdf.plot.kde_2d('x', 'y', style='c')
    with pytest.raises(TypeError):
        ax = mcmc_wdf.plot.scatter_2d('x', 'y', style='c')



from typing import Literal
import numpy as np
from matplotlib import pyplot as plt
from pandas.core.dtypes.generic import ABCMultiIndex
import pandas.core.common as com
from pandas.plotting._matplotlib.core import (ScatterPlot as _ScatterPlot,
                                              HexBinPlot as _HexBinPlot,
                                              LinePlot as _LinePlot,
                                              BarPlot as _BarPlot,
                                              BarhPlot as _BarhPlot,
                                              AreaPlot as _AreaPlot,
                                              PiePlot as _PiePlot,
                                              MPLPlot, PlanePlot)
from pandas.plotting._matplotlib.groupby import create_iter_data_given_by
from pandas.io.formats.printing import pprint_thing
from anesthetic.weighted_pandas import _WeightedObject
from anesthetic.plot import scatter_plot_2d


def _get_weights(kwargs, data):
    if isinstance(data, _WeightedObject):
        kwargs['weights'] = data.get_weights()


class _WeightedMPLPlot(MPLPlot):

    _default_rot = None

    def __init__(self, data, *args, **kwargs):
        _get_weights(kwargs, data)
        super().__init__(data, *args, **kwargs)

    def _get_index_name(self):
        if isinstance(self.data, _WeightedObject):
            if isinstance(self.data.drop_weights().index, ABCMultiIndex):
                name = self.data.drop_weights().index.names
                if com.any_not_none(*name):
                    name = ",".join([pprint_thing(x) for x in name])
                else:
                    name = None
            else:
                name = self.data.drop_weights().index.name
                if name is not None:
                    name = pprint_thing(name)

            # GH 45145, override the default axis label if one is provided.
            index_name = self._get_custom_index_name()
            if index_name is not None:
                name = pprint_thing(index_name)

            return name
        else:
            return super()._get_index_name()

    def _get_xticks(self, convert_period: bool = False):
        if isinstance(self.data, _WeightedObject):
            return self.data.drop_weights().index._mpl_repr()
        else:
            return super()._get_xticks(convert_period)


def _compress_weights(kwargs, data):
    if isinstance(data, _WeightedObject):
        return data.compress(kwargs.pop('ncompress', True))
    else:
        return data


class _CompressedMPLPlot(MPLPlot):
    def __init__(self, data, *args, **kwargs):
        data = _compress_weights(kwargs, data)
        super().__init__(data, *args, **kwargs)


class ScatterPlot(_CompressedMPLPlot, _ScatterPlot):
    # noqa: disable=D101
    def __init__(self, data, x, y, s=None, c=None, **kwargs) -> None:
        if isinstance(data, _WeightedObject):
            kwargs['alpha'] = kwargs.get('alpha', 0.5)
        super().__init__(data, x, y, s, c, **kwargs)


class _PlanePlot2d(PlanePlot):

    def _make_plot(self):
        if self.colormap is not None:
            self.kwds['cmap'] = plt.get_cmap(self.colormap)
        colors = self._get_colors()
        data = (
            create_iter_data_given_by(self.data, self.kind)  # safe for now
            if self.by is not None
            else self.data
        )
        x = data[self.x].values
        y = data[self.y].values
        ax = self._get_ax(0)  # another one of these hard-coded 0s

        kwds = self.kwds.copy()
        label = pprint_thing(self.label)
        kwds["label"] = label

        style, kwds = self._apply_style_colors(colors, kwds, 0, label)
        if style is not None:
            raise TypeError("'style' keyword argument is not "
                            f"supported by {self._kind}")
        self._plot(ax, x, y, **kwds)

    def _args_adjust(self):
        pass


class ScatterPlot2d(_CompressedMPLPlot, _PlanePlot2d):
    # noqa: disable=D101
    @property
    def _kind(self) -> Literal["scatter_2d"]:
        return "scatter_2d"

    @classmethod
    def _plot(cls, ax, x, y, **kwds):
        return scatter_plot_2d(ax, x, y, **kwds)


class HexBinPlot(_HexBinPlot):
    # noqa: disable=D101
    def __init__(self, data, x, y, C=None, **kwargs) -> None:
        if isinstance(data, _WeightedObject):
            C = '__weights'
            data[C] = data.get_weights()
            kwargs['reduce_C_function'] = np.sum
        super().__init__(data, x, y, C=C, **kwargs)


class LinePlot(_LinePlot, _WeightedMPLPlot):
    # noqa: disable=D101
    pass


class PiePlot(_PiePlot):
    # noqa: disable=D101
    def __init__(self, data, kind=None, **kwargs) -> None:
        if isinstance(data, _WeightedObject):
            labels = data.drop_weights().index._mpl_repr()
            kwargs['labels'] = kwargs.get('labels', labels)
        super().__init__(data, kind=kind, **kwargs)


class BarPlot(_BarPlot, _WeightedMPLPlot):
    # noqa: disable=D101
    def _decorate_ticks(self, ax, name, ticklabels, start_edge, end_edge):
        super()._decorate_ticks(ax, name, ticklabels, start_edge, end_edge)
        if isinstance(self.data, _WeightedObject):
            if self.xticks is None:
                ax.set_xticklabels(self._get_xticks())


class BarhPlot(_BarhPlot, _WeightedMPLPlot):
    # noqa: disable=D101
    def _decorate_ticks(self, ax, name, ticklabels, start_edge, end_edge):
        super()._decorate_ticks(ax, name, ticklabels, start_edge, end_edge)
        if isinstance(self.data, _WeightedObject):
            if self.yticks is None:
                ax.set_yticklabels(self._get_xticks())


class AreaPlot(_AreaPlot, _WeightedMPLPlot):
    # noqa: disable=D101
    pass



"""Read NestedSamples from multinest chains."""
import os
import numpy as np
from anesthetic.read.getdist import read_paramnames
from anesthetic.samples import NestedSamples


def read_multinest(root, *args, **kwargs):
    """Read <root>ev.dat and <root>phys_live.points in MultiNest format."""
    try:
        data = np.loadtxt(root + 'dead-birth.txt')
        samples, logL, logL_birth, _ = np.split(data, [-4, -3, -2], axis=1)
        logL = np.squeeze(logL)
        logL_birth = np.squeeze(logL_birth)

        data = np.loadtxt(root + 'phys_live-birth.txt')
        (live_samples, live_logL,
         live_logL_birth, _) = np.split(data, [-3, -2, -1], axis=1)
        live_logL = np.squeeze(live_logL)
        live_logL_birth = np.squeeze(live_logL_birth)
        i = np.argsort(live_logL)
        samples = np.concatenate((samples, live_samples[i]), axis=0)
        logL = np.concatenate((logL, live_logL[i]))
        logL_birth = np.concatenate((logL_birth, live_logL_birth[i]))

    except (FileNotFoundError, IOError):
        data = np.loadtxt(root + 'ev.dat')
        samples, logL, _ = np.split(data, [-3, -2], axis=1)
        logL = np.squeeze(logL)

        data = np.loadtxt(root + 'phys_live.points')
        live_samples, live_logL, _ = np.split(data, [-2, -1], axis=1)
        live_logL = np.squeeze(live_logL)
        i = np.argsort(live_logL)
        logL_birth = len(live_logL)
        samples = np.concatenate((samples, live_samples[i]), axis=0)
        logL = np.concatenate((logL, live_logL[i]))

    kwargs['label'] = kwargs.get('label', os.path.basename(root))
    columns, labels = read_paramnames(root)
    columns = kwargs.pop('columns', columns)
    labels = kwargs.pop('labels', labels)
    data = samples

    return NestedSamples(data=data, logL=logL, logL_birth=logL_birth,
                         columns=columns, labels=labels, *args, **kwargs)



from anesthetic import read_chains
from anesthetic.convert import to_getdist
from numpy.testing import assert_array_equal


def test_to_getdist():
    try:
        anesthetic_samples = read_chains('./tests/example_data/gd')
        getdist_samples = to_getdist(anesthetic_samples)

        assert_array_equal(getdist_samples.samples, anesthetic_samples)
        assert_array_equal(getdist_samples.weights,
                           anesthetic_samples.get_weights())

        for param, p in zip(getdist_samples.getParamNames().names,
                            anesthetic_samples.drop_labels().columns):

            assert param.name == p
    except ModuleNotFoundError:
        pass



import anesthetic.examples._matplotlib_agg  # noqa: F401
from anesthetic import read_chains
import pytest
import pandas._testing as tm


@pytest.fixture(autouse=True)
def close_figures_on_teardown():
    tm.close()


def test_gui():
    plotter = read_chains('./tests/example_data/pc').gui()

    # Type buttons
    plotter.type.buttons.set_active(1)
    assert plotter.type() == 'posterior'
    plotter.type.buttons.set_active(0)
    assert plotter.type() == 'live'

    # Parameter choice buttons
    plotter.param_choice.buttons.set_active(1)
    assert len(plotter.triangle.ax) == 2
    plotter.param_choice.buttons.set_active(0)
    assert len(plotter.triangle.ax) == 1
    plotter.param_choice.buttons.set_active(0)
    plotter.param_choice.buttons.set_active(2)
    plotter.param_choice.buttons.set_active(3)
    assert len(plotter.triangle.ax) == 4

    # Sliders
    plotter.evolution.slider.set_val(100)
    assert plotter.evolution() == 100
    plotter.type.buttons.set_active(1)

    plotter.temperature.slider.set_val(0)
    assert plotter.temperature() == 1
    plotter.temperature.slider.set_val(1)
    assert plotter.temperature() == 10
    plotter.temperature.slider.set_val(2)
    assert plotter.temperature() == 100
    plotter.type.buttons.set_active(0)

    # Reload button
    plotter.reload.button.on_clicked(plotter.reload_file(None))

    # Reset button
    plotter.reset.button.on_clicked(plotter.reset_range(None))


def test_gui_params():
    plotter = read_chains('./tests/example_data/pc').gui()
    assert len(plotter.param_choice.buttons.labels) == 8

    plotter = read_chains('./tests/example_data/pc').gui(params=['x0', 'x1'])
    assert len(plotter.param_choice.buttons.labels) == 2


def test_slider_reset_range():
    plotter = read_chains('./tests/example_data/pc').gui()
    plotter.evolution.reset_range(-3, 3)
    assert plotter.evolution.ax.get_xlim() == (-3.0, 3.0)



import anesthetic.examples._matplotlib_agg  # noqa: F401
import pytest
import numpy as np
import sys
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
from anesthetic.plot import (make_1d_axes, make_2d_axes, kde_plot_1d,
                             fastkde_plot_1d, hist_plot_1d, hist_plot_2d,
                             fastkde_contour_plot_2d, kde_contour_plot_2d,
                             scatter_plot_2d, quantile_plot_interval,
                             basic_cmap, AxesSeries, AxesDataFrame)
from numpy.testing import assert_array_equal
import pandas._testing as tm

from matplotlib.axes import SubplotBase
from matplotlib.contour import ContourSet
from matplotlib.lines import Line2D
from matplotlib.patches import Patch, Polygon
from matplotlib.colors import ColorConverter, to_rgba
from matplotlib.figure import Figure
from scipy.special import erf
from scipy.interpolate import interp1d


@pytest.fixture(autouse=True)
def close_figures_on_teardown():
    tm.close()


def test_AxesObjects():
    paramnames = ['a', 'b', 'c']

    # AxesSeries
    axes = AxesSeries(index=paramnames)
    assert isinstance(axes, AxesSeries)
    assert isinstance(axes.iloc[0], SubplotBase)
    assert axes.iloc[0].get_xlabel() == 'a'
    axes.set_xlabels(labels=dict(a='A', b='B', c='C'))
    assert axes.iloc[0].get_xlabel() == 'A'
    axes.tick_params(labelrotation=0, labelsize='medium')

    # AxesDataFrame
    axes = AxesDataFrame(index=paramnames + ['d'], columns=paramnames)
    assert isinstance(axes, AxesDataFrame)
    assert isinstance(axes.iloc[0, 0], SubplotBase)
    assert axes.iloc[-1, 0].get_xlabel() == 'a'
    assert axes.iloc[-1, 0].get_ylabel() == 'd'
    axes.set_labels(labels=dict(a='A', b='B', c='C', d='D'))
    assert axes.iloc[-1, 0].get_xlabel() == 'A'
    assert axes.iloc[-1, 0].get_ylabel() == 'D'
    axes.tick_params(labelrotation=0, labelsize='medium')
    xmin1, xmax1 = axes.iloc[0, 0].get_xlim()
    ymin1, ymax1 = axes.iloc[0, 0].get_ylim()
    axes.set_margins(m=0.5)
    xmin2, xmax2 = axes.iloc[0, 0].get_xlim()
    ymin2, ymax2 = axes.iloc[0, 0].get_ylim()
    assert xmin2 == xmin1 - 0.5 * (xmax1 - xmin1)
    assert xmax2 == xmax1 + 0.5 * (xmax1 - xmin1)
    assert ymin2 == ymin1 - 0.5 * (ymax1 - ymin1)
    assert ymax2 == ymax1 + 0.5 * (ymax1 - ymin1)


def test_make_1d_axes():
    paramnames = ['A', 'B', 'C', 'D', 'E']
    labels = {'A': 'tA', 'B': 'tB', 'C': 'tC', 'D': 'tD', 'E': 'tE'}

    # Check no optional arguments
    fig, axes = make_1d_axes(paramnames)
    assert isinstance(fig, Figure)
    assert isinstance(axes, AxesSeries)
    assert isinstance(axes.to_frame(), AxesDataFrame)
    assert_array_equal(axes.index, paramnames)
    for p, ax in axes.items():
        assert ax.get_xlabel() == p

    # Check single string input
    fig, axes = make_1d_axes(paramnames[0])
    assert isinstance(fig, Figure)
    assert isinstance(axes, AxesSeries)
    assert axes.index.size == 1
    assert_array_equal(axes.index, paramnames[0])

    # Check labels argument
    fig, axes = make_1d_axes(paramnames, labels=labels)
    for t in labels:
        assert axes[t].get_xlabel() != t
        assert axes[t].get_xlabel() == labels[t]

    # Check fig argument
    fig0 = plt.figure()
    fig0.suptitle('hi there')
    fig, axes = make_1d_axes(paramnames)
    assert fig is not fig0
    fig, axes = make_1d_axes(paramnames, fig=fig0)
    assert fig is fig0

    # Check ncol argument
    fig, axes = make_1d_axes(paramnames, ncol=2)
    nrows, ncol = axes[0].get_subplotspec().get_gridspec().get_geometry()
    assert ncol == 2

    # Check gridspec argument
    grid = gs.GridSpec(2, 2, width_ratios=[3, 1], height_ratios=[3, 1])
    g00 = grid[0, 0]
    fig, axes = make_1d_axes(paramnames, subplot_spec=g00)
    assert g00 is axes[0].get_subplotspec().get_topmost_subplotspec()

    # Check gridspec kwargs
    fig, axes = make_1d_axes(paramnames, gridspec_kw=dict(wspace=0.1))
    ws = axes[0].get_subplotspec().get_gridspec().get_subplot_params().wspace
    assert ws == 0.1

    # Check figure kwargs
    fig, axes = make_1d_axes(paramnames, figsize=(5, 5))
    assert fig.get_figwidth() == 5
    assert fig.get_figheight() == 5

    # Check unexpected kwargs
    with pytest.raises((AttributeError, TypeError)):
        make_1d_axes(paramnames, spam='ham')


def test_make_2d_axes_inputs_outputs():
    paramnames_x = ['A', 'B', 'C', 'D']
    paramnames_y = ['B', 'A', 'D', 'E']

    # 2D axes
    fig, axes = make_2d_axes([paramnames_x, paramnames_y])
    assert isinstance(fig, Figure)
    assert isinstance(axes, AxesDataFrame)
    assert isinstance(axes['A'], AxesSeries)
    assert isinstance(axes.loc['A':'B', 'B':'C'], AxesDataFrame)
    assert_array_equal(axes.index, paramnames_y)
    assert_array_equal(axes.columns, paramnames_x)

    # Axes labels
    for p, ax in axes.iloc[:, 0].items():
        assert ax.get_ylabel() == p

    for p, ax in axes.iloc[-1].items():
        assert ax.get_xlabel() == p

    for ax in axes.iloc[:-1, 1:].to_numpy().flatten():
        assert ax.get_xlabel() == ''
        assert ax.get_ylabel() == ''

    # Check fig argument
    fig0 = plt.figure()
    fig, axes = make_2d_axes(paramnames_x)
    assert fig is not fig0
    fig, axes = make_2d_axes(paramnames_x, fig=fig0)
    assert fig is fig0

    # Check gridspec argument
    grid = gs.GridSpec(2, 2, width_ratios=[3, 1], height_ratios=[3, 1])
    g00 = grid[0, 0]
    fig, axes = make_2d_axes(paramnames_x, subplot_spec=g00)
    assert g00 is axes.iloc[0, 0].get_subplotspec().get_topmost_subplotspec()

    # Check gridspec kwargs
    fig, axes = make_1d_axes(paramnames_x, gridspec_kw=dict(wspace=0.1))
    ws = axes[0].get_subplotspec().get_gridspec().get_subplot_params().wspace
    assert ws == 0.1

    # Check figure kwargs
    fig, axes = make_2d_axes(paramnames_x, figsize=(5, 5))
    assert fig.get_figwidth() == 5
    assert fig.get_figheight() == 5

    # Check unexpected kwargs
    with pytest.raises((AttributeError, TypeError)):
        make_2d_axes(paramnames_x, spam='ham')


@pytest.mark.parametrize('paramnames_y', [['A', 'B', 'C', 'D'],
                                          ['A', 'C', 'B', 'D'],
                                          ['D', 'C', 'B', 'A'],
                                          ['C', 'B', 'A'],
                                          ['E', 'F', 'G', 'H'],
                                          ['A', 'B', 'E', 'F'],
                                          ['B', 'E', 'A', 'F'],
                                          ['B', 'F', 'A', 'H', 'G'],
                                          ['B', 'A', 'H', 'G']])
@pytest.mark.parametrize('upper', [False, True])
@pytest.mark.parametrize('lower', [False, True])
@pytest.mark.parametrize('diagonal', [False, True])
def test_make_2d_axes_behaviour(diagonal, lower, upper, paramnames_y):
    np.random.seed(0)

    def calc_n(axes):
        """Compute the number of upper, lower and diagonal plots."""
        n = {'upper': 0, 'lower': 0, 'diagonal': 0}
        for y, row in axes.iterrows():
            for x, ax in row.items():
                if ax is not None:
                    n[ax.position] += 1
        return n

    # Check for only paramnames_x
    paramnames_x = ['A', 'B', 'C', 'D']
    nx = len(paramnames_x)
    fig, axes = make_2d_axes(paramnames_x,
                             upper=upper,
                             lower=lower,
                             diagonal=diagonal)
    ns = calc_n(axes)
    assert ns['upper'] == upper * nx*(nx-1)//2
    assert ns['lower'] == lower * nx*(nx-1)//2
    assert ns['diagonal'] == diagonal * nx

    params = [paramnames_x, paramnames_y]
    all_params = paramnames_x + paramnames_y
    nu, nl, nd = 0, 0, 0
    for x in paramnames_x:
        for y in paramnames_y:
            if x == y:
                nd += 1
            elif all_params.index(x) < all_params.index(y):
                nl += 1
            elif all_params.index(x) > all_params.index(y):
                nu += 1
    fig, axes = make_2d_axes(params,
                             upper=upper,
                             lower=lower,
                             diagonal=diagonal)
    ns = calc_n(axes)
    assert ns['upper'] == upper * nu
    assert ns['lower'] == lower * nl
    assert ns['diagonal'] == diagonal * nd


@pytest.mark.parametrize('upper', [False, True])
@pytest.mark.parametrize('ticks', ['inner', 'outer', None])
def test_make_2d_axes_ticks(upper, ticks):
    xticks = [0.1, 0.4, 0.7]
    yticks = [0.2, 0.5, 0.8]
    paramnames = ["x0", "x1", "x2", "x3"]
    for k in paramnames:
        fig, axes = make_2d_axes(paramnames, upper=upper, ticks=ticks)
        axes[k][k].set_xticks(xticks)
        axes[k][k].set_yticks(yticks)
        for i, row in axes.iterrows():
            for j, ax in row.items():
                if ax is None:
                    break
                if i == k:
                    assert np.array_equal(yticks, ax.get_yticks())
                else:
                    assert not np.array_equal(yticks, ax.get_yticks())
                if j == k:
                    assert np.array_equal(xticks, ax.get_xticks())
                else:
                    assert not np.array_equal(xticks, ax.get_xticks())
    with pytest.raises(ValueError):
        make_2d_axes(paramnames, upper=upper, ticks='spam')


def test_make_2d_axes_ticks_error():
    with pytest.raises(ValueError):
        make_2d_axes(['a', 'b'], ticks='spam')


def test_2d_axes_limits():
    np.random.seed(0)
    paramnames = ['A', 'B', 'C', 'D']
    fig, axes = make_2d_axes(paramnames)
    for x in paramnames:
        for y in paramnames:
            a, b, c, d = np.random.rand(4)
            axes[x][y].set_xlim(a, b)
            for z in paramnames:
                assert axes[x][z].get_xlim() == (a, b)
                assert axes[z][x].get_ylim() == (a, b)

            axes[x][y].set_ylim(c, d)
            for z in paramnames:
                assert axes[y][z].get_xlim() == (c, d)
                assert axes[z][y].get_ylim() == (c, d)


@pytest.mark.parametrize('axesparams', [['A', 'B', 'C', 'D'],
                                        [['A', 'B', 'C', 'D'], ['A', 'B']],
                                        [['A', 'B'], ['A', 'B', 'C', 'D']]])
@pytest.mark.parametrize('params', [{'A': 0},
                                    {'A': 0, 'C': 0, 'E': 0},
                                    {'A': 0, 'C': [0, 0.5]}])
@pytest.mark.parametrize('upper', [True, False])
def test_2d_axes_axlines(axesparams, params, upper):
    kwargs = dict(c='k', ls='--', lw=0.5)
    fig, axes = make_2d_axes(axesparams, upper=upper)
    axes.axlines(params, **kwargs)


@pytest.mark.parametrize('axesparams', [['A', 'B', 'C', 'D'],
                                        [['A', 'B', 'C', 'D'], ['A', 'B']],
                                        [['A', 'B'], ['A', 'B', 'C', 'D']]])
@pytest.mark.parametrize('params', [{'A': (0, 0.1)},
                                    {'A': (0, 1), 'C': (0, 1), 'E': (0, 1)},
                                    {'A': (0, 1), 'C': [(-0.1, 0), (0.5, 1)]}])
@pytest.mark.parametrize('upper', [True, False])
def test_2d_axes_axspans(axesparams, params, upper):
    kwargs = dict(c='k', alpha=0.5)
    fig, axes = make_2d_axes(axesparams, upper=upper)
    axes.axspans(params, **kwargs)


@pytest.mark.parametrize('axesparams', [['A', 'B', 'C', 'D'],
                                        [['A', 'B', 'C', 'D'], ['A', 'B']],
                                        [['A', 'B'], ['A', 'B', 'C', 'D']]])
@pytest.mark.parametrize('params', [{'A': 0},
                                    {'A': 0, 'C': 0, 'E': 0},
                                    {'A': [0, 0.1], 'C': [0, 0.5]}])
@pytest.mark.parametrize('upper', [True, False])
def test_2d_axes_scatter(axesparams, params, upper):
    kwargs = dict(c='k', marker='*')
    fig, axes = make_2d_axes(axesparams, upper=upper)
    axes.scatter(params, **kwargs)


@pytest.mark.parametrize('plot_1d', [kde_plot_1d, fastkde_plot_1d])
def test_kde_plot_1d(plot_1d):
    fig, ax = plt.subplots()
    np.random.seed(0)
    data = np.random.randn(1000)

    try:
        # Check height
        line, = plot_1d(ax, data)
        assert isinstance(line, Line2D)
        assert line.get_ydata().max() <= 1

        # Check arguments are passed onward to underlying function
        line, = plot_1d(ax, data, color='r')
        assert line.get_color() == 'r'
        line, = plot_1d(ax, data, cmap=plt.cm.Blues)
        assert line.get_color() == plt.cm.Blues(0.68)

        # Check q
        plot_1d(ax, data, q='1sigma')
        plot_1d(ax, data, q=0)
        plot_1d(ax, data, q=1)
        plot_1d(ax, data, q=5)
        plot_1d(ax, data, q=10)
        plot_1d(ax, data, q=0.1)
        plot_1d(ax, data, q=0.9)
        plot_1d(ax, data, q=(0.1, 0.9))

        # Check iso-probability code
        line, fill = plot_1d(ax, data, facecolor=True)
        plot_1d(ax, data, facecolor=True, levels=[0.8, 0.6, 0.2])
        line, fill = plot_1d(ax, data, fc='blue', color='k', ec='r')
        assert np.all(fill[0].get_edgecolor()[0] == to_rgba('r'))
        assert (to_rgba(line[0].get_color()) == to_rgba('r'))
        line, fill = plot_1d(ax, data, fc=True, color='k', ec=None)
        assert len(fill[0].get_edgecolor()) == 0
        assert (to_rgba(line[0].get_color()) == to_rgba('k'))

        # Check levels
        with pytest.raises(ValueError):
            fig, ax = plt.subplots()
            plot_1d(ax, data, fc=True, levels=[0.68, 0.95])

        # Check xlim, Gaussian (i.e. limits reduced to relevant data range)
        fig, ax = plt.subplots()
        data = np.random.randn(1000) * 0.01 + 0.5
        plot_1d(ax, data)
        xmin, xmax = ax.get_xlim()
        assert xmin > 0.4
        assert xmax < 0.6
        # Check xlim, Uniform (i.e. data and limits span entire prior boundary)
        fig, ax = plt.subplots()
        data = np.random.uniform(size=1000)
        plot_1d(ax, data, q=0)
        xmin, xmax = ax.get_xlim()
        assert xmin <= 0
        assert xmax >= 1

    except ImportError:
        if 'fastkde' not in sys.modules:
            pass


def test_fastkde_min_max():
    np.random.seed(0)
    data_x = np.random.randn(100)
    data_y = np.random.randn(100)
    xmin, xmax = -1, +1
    ymin, ymax = -1, +1
    try:
        _, ax = plt.subplots()
        line, = fastkde_plot_1d(ax, data_x, xmin=xmin)
        assert (line.get_xdata() >= xmin).all()

        _, ax = plt.subplots()
        line, = fastkde_plot_1d(ax, data_x, xmax=xmax)
        assert (line.get_xdata() <= xmax).all()

        _, ax = plt.subplots()
        line, = fastkde_plot_1d(ax, data_x, xmin=xmin, xmax=xmax)
        assert (line.get_xdata() >= xmin).all()
        assert (line.get_xdata() <= xmax).all()

        _, ax = plt.subplots()
        fastkde_contour_plot_2d(ax, data_x, data_y, xmin=xmin, ymin=ymin)
        assert ax.get_xlim()[0] >= xmin
        assert ax.get_ylim()[0] >= ymin

        _, ax = plt.subplots()
        fastkde_contour_plot_2d(ax, data_x, data_y, xmax=xmax, ymax=ymax)
        assert ax.get_xlim()[1] <= xmax
        assert ax.get_ylim()[1] <= ymax

        _, ax = plt.subplots()
        fastkde_contour_plot_2d(ax, data_x, data_y,
                                xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)
        assert ax.get_xlim()[0] >= xmin
        assert ax.get_xlim()[1] <= xmax
        assert ax.get_ylim()[0] >= ymin
        assert ax.get_ylim()[1] <= ymax

    except ImportError:
        if 'fastkde' not in sys.modules:
            pass


def test_hist_plot_1d():
    fig, ax = plt.subplots()
    np.random.seed(0)
    data = np.random.randn(1000)

    # Check heights for histtype 'bar'
    bars = hist_plot_1d(ax, data, histtype='bar')[2]
    assert np.all([isinstance(b, Patch) for b in bars])
    assert max([b.get_height() for b in bars]) == 1.
    assert np.all(np.array([b.get_height() for b in bars]) <= 1.)

    # Check heights for histtype 'step'
    polygon, = hist_plot_1d(ax, data, histtype='step')[2]
    assert isinstance(polygon, Polygon)
    trans = polygon.get_transform() - ax.transData
    assert np.isclose(trans.transform(polygon.xy)[:, -1].max(), 1.,
                      rtol=1e-10, atol=1e-10)
    assert np.all(trans.transform(polygon.xy)[:, -1] <= 1.)

    # Check heights for histtype 'stepfilled'
    polygon, = hist_plot_1d(ax, data, histtype='stepfilled')[2]
    assert isinstance(polygon, Polygon)
    trans = polygon.get_transform() - ax.transData
    assert np.isclose(trans.transform(polygon.xy)[:, -1].max(), 1.,
                      rtol=1e-10, atol=1e-10)
    assert np.all(trans.transform(polygon.xy)[:, -1] <= 1.)

    # Check arguments are passed onward to underlying function
    bars = hist_plot_1d(ax, data, histtype='bar', color='r', alpha=0.5)[2]
    cc = ColorConverter.to_rgba('r', alpha=0.5)
    assert np.all([b.get_fc() == cc for b in bars])
    bars = hist_plot_1d(ax, data, histtype='bar', cmap=plt.cm.viridis,
                        alpha=0.5)[2]
    cc = ColorConverter.to_rgba(plt.cm.viridis(0.68), alpha=0.5)
    assert np.all([b.get_fc() == cc for b in bars])
    polygon, = hist_plot_1d(ax, data, histtype='step', color='r', alpha=0.5)[2]
    assert polygon.get_ec() == ColorConverter.to_rgba('r', alpha=0.5)
    polygon, = hist_plot_1d(ax, data, histtype='step', cmap=plt.cm.viridis,
                            color='r')[2]
    assert polygon.get_ec() == ColorConverter.to_rgba('r')


@pytest.mark.parametrize('bins', ['knuth', 'freedman', 'blocks'])
def test_astropyhist_plot_1d(bins):
    try:
        fig, ax = plt.subplots()
        np.random.seed(0)
        data = np.random.randn(100)
        bars = hist_plot_1d(ax, data, histtype='bar', bins=bins)[2]
        assert np.all([isinstance(b, Patch) for b in bars])
        assert max([b.get_height() for b in bars]) == 1.
        assert np.all(np.array([b.get_height() for b in bars]) <= 1.)
    except ImportError:
        if 'astropy' not in sys.modules:
            pass


def test_hist_plot_2d():
    fig, ax = plt.subplots()
    np.random.seed(0)
    data_x, data_y = np.random.randn(2, 1000)
    hist_plot_2d(ax, data_x, data_y)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    assert xmin > -5 and xmax < 5 and ymin > -5 and ymax < 5

    data_x, data_y = np.random.uniform(-10, 10, (2, 1000))
    hist_plot_2d(ax, data_x, data_y)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    assert xmin > -10 and xmax < 10 and ymin > -10 and ymax < 10

    data_x, data_y = np.random.uniform(-10, 10, (2, 1000))
    weights = np.exp(-(data_x**2 + data_y**2)/2)
    hist_plot_2d(ax, data_x, data_y, weights=weights, bins=30)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    assert xmin > -6 and xmax < 6 and ymin > -6 and ymax < 6

    fig, ax = plt.subplots()
    data_x, data_y = np.random.randn(2, 1000)
    hist_plot_2d(ax, data_x, data_y, levels=[0.95, 0.68], cmin=50)
    hist_plot_2d(ax, data_x, data_y, levels=[0.95, 0.68], cmax=50)


@pytest.mark.parametrize('plot_1d', [kde_plot_1d, fastkde_plot_1d])
@pytest.mark.parametrize('s', [1, 2])
def test_1d_density_kwarg(plot_1d, s):
    try:
        np.random.seed(0)
        x = np.random.normal(scale=s, size=2000)
        fig, ax = plt.subplots()

        # hist density = False:
        h = hist_plot_1d(ax, x, density=False,
                         bins=np.linspace(-5.5, 5.5, 12))[2]
        bar_height = h.get_children()[len(h.get_children()) // 2].get_height()
        assert bar_height == pytest.approx(1, rel=0.1)

        # kde density = False:
        k = plot_1d(ax, x, density=False)[0]
        f = interp1d(k.get_xdata(), k.get_ydata(), 'cubic', assume_sorted=True)
        kde_height = f(0)
        assert kde_height == pytest.approx(1, rel=0.1)

        # hist density = True:
        h = hist_plot_1d(ax, x, density=True,
                         bins=np.linspace(-5.5, 5.5, 12))[2]
        bar_height = h.get_children()[len(h.get_children()) // 2].get_height()
        assert bar_height == pytest.approx(erf(0.5 / np.sqrt(2) / s), rel=0.1)

        # kde density = True:
        k = plot_1d(ax, x, density=True)[0]
        f = interp1d(k.get_xdata(), k.get_ydata(), 'cubic', assume_sorted=True)
        kde_height = f(0)
        gauss_norm = 1 / np.sqrt(2 * np.pi * s**2)
        assert kde_height == pytest.approx(gauss_norm, rel=0.1)

    except ImportError:
        if 'fastkde' not in sys.modules:
            pass


@pytest.mark.parametrize('contour_plot_2d', [kde_contour_plot_2d,
                                             fastkde_contour_plot_2d])
def test_contour_plot_2d(contour_plot_2d):
    try:
        fig, ax = plt.subplots()
        np.random.seed(1)
        data_x = np.random.randn(1000)
        data_y = np.random.randn(1000)
        cf, ct = contour_plot_2d(ax, data_x, data_y)
        assert isinstance(cf, ContourSet)
        assert isinstance(ct, ContourSet)

        # Check levels
        with pytest.raises(ValueError):
            fig, ax = plt.subplots()
            contour_plot_2d(ax, data_x, data_y, levels=[0.68, 0.95])

        # Check q
        fig, ax = plt.subplots()
        contour_plot_2d(ax, data_x, data_y, q=0)

        # Check unfilled
        cmap = basic_cmap('C2')
        fig, ax = plt.subplots()
        cf1, ct1 = contour_plot_2d(ax, data_x, data_y, facecolor='C2')
        cf2, ct2 = contour_plot_2d(ax, data_x, data_y, fc='None', cmap=cmap)
        # filled `contourf` and unfilled `contour` colors are the same:
        assert cf1.tcolors[0] == ct2.tcolors[0]
        assert cf1.tcolors[1] == ct2.tcolors[1]
        cf, ct = contour_plot_2d(ax, data_x, data_y, edgecolor='C0')
        assert ct.colors == 'C0'
        cf, ct = contour_plot_2d(ax, data_x, data_y, ec='C0', cmap=plt.cm.Reds)
        assert cf.get_cmap() == plt.cm.Reds
        assert ct.colors == 'C0'
        fig, ax = plt.subplots()
        cf, ct = contour_plot_2d(ax, data_x, data_y, fc=None)
        assert cf is None
        assert ct.colors is None
        assert ct.get_cmap()(1.) == to_rgba('C0')
        cf, ct = contour_plot_2d(ax, data_x, data_y, fc=None, c='C3')
        assert cf is None
        assert ct.colors is None
        assert ct.get_cmap()(1.) == to_rgba('C3')
        cf, ct = contour_plot_2d(ax, data_x, data_y, fc=None, ec='C1')
        assert cf is None
        assert ct.colors == 'C1'
        cf, ct = contour_plot_2d(ax, data_x, data_y, fc=None, cmap=plt.cm.Reds)
        assert cf is None
        assert ct.colors is None
        assert ct.get_cmap() == plt.cm.Reds

        # Check limits, Gaussian (i.e. limits reduced to relevant data range)
        fig, ax = plt.subplots()
        data_x = np.random.randn(1000) * 0.01 + 0.5
        data_y = np.random.randn(1000) * 0.01 + 0.5
        contour_plot_2d(ax, data_x, data_y)
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        assert xmin > 0.4
        assert xmax < 0.6
        assert ymin > 0.4
        assert ymax < 0.6
        # Check limits, Uniform (i.e. data & limits span entire prior boundary)
        fig, ax = plt.subplots()
        data_x = np.random.uniform(size=1000)
        data_y = np.random.uniform(size=1000)
        contour_plot_2d(ax, data_x, data_y, q=0)
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        if contour_plot_2d is fastkde_contour_plot_2d:
            assert xmin <= 0
            assert xmax >= 1
            assert ymin <= 0
            assert ymax >= 1
        elif contour_plot_2d is kde_contour_plot_2d:
            assert xmin == pytest.approx(0, abs=0.01)
            assert xmax == pytest.approx(1, abs=0.01)
            assert ymin == pytest.approx(0, abs=0.01)
            assert ymax == pytest.approx(1, abs=0.01)

    except ImportError:
        if 'fastkde' not in sys.modules:
            pass


def test_kde_plot_nplot():
    fig, ax = plt.subplots()
    np.random.seed(0)
    data = np.random.randn(1000)
    line, = kde_plot_1d(ax, data, ncompress=1000, nplot_1d=200)
    assert line.get_xdata().size == 200

    fig, ax = plt.subplots()
    np.random.seed(0)
    data_x = np.random.randn(1000)
    data_y = np.random.randn(1000)
    kde_contour_plot_2d(ax, data_x, data_y, ncompress=1000, nplot_2d=900)


@pytest.mark.parametrize('contour_plot_2d', [kde_contour_plot_2d,
                                             fastkde_contour_plot_2d])
@pytest.mark.parametrize('levels', [[0.9],
                                    [0.9, 0.6],
                                    [0.9, 0.6, 0.3],
                                    [0.9, 0.7, 0.5, 0.3]])
def test_contour_plot_2d_levels(contour_plot_2d, levels):
    try:
        np.random.seed(42)
        x = np.random.randn(1000)
        y = np.random.randn(1000)
        cmap = plt.cm.viridis

        fig, (ax1, ax2) = plt.subplots(2)
        contour_plot_2d(ax1, x, y, levels=levels, cmap=cmap)
        contour_plot_2d(ax2, x, y, levels=levels, cmap=cmap, fc=None)

        # assert that color between filled and unfilled contours matches
        # first level
        color1 = ax1.collections[0].get_facecolor()  # filled face color
        color2 = ax2.collections[0].get_edgecolor()  # unfilled line color
        assert_array_equal(color1, color2)
        # last level
        color1 = ax1.collections[len(levels)-1].get_facecolor()
        color2 = ax2.collections[len(levels)-1].get_edgecolor()
        assert_array_equal(color1, color2)

    except ImportError:
        if 'fastkde' not in sys.modules:
            pass


def test_scatter_plot_2d():
    fig, ax = plt.subplots()
    np.random.seed(2)
    data_x = np.random.randn(100)
    data_y = np.random.randn(100)
    lines, = scatter_plot_2d(ax, data_x, data_y)
    assert isinstance(lines, Line2D)

    fig, ax = plt.subplots()
    points, = scatter_plot_2d(ax, data_x, data_y, color='C0', lw=1)
    assert (points.get_color() == 'C0')
    points, = scatter_plot_2d(ax, data_x, data_y, cmap=plt.cm.viridis)
    assert (points.get_color() == plt.cm.viridis(0.68))
    points, = scatter_plot_2d(ax, data_x, data_y, c='C0', fc='C1', ec='C2')
    assert (points.get_color() == 'C0')
    assert (points.get_markerfacecolor() == 'C1')
    assert (points.get_markeredgecolor() == 'C2')

    # Check that q is ignored
    fig, ax = plt.subplots()
    scatter_plot_2d(ax, data_x, data_y, q=0)


@pytest.mark.parametrize('sigmas', [(1, '1sigma', 0.682689492137086),
                                    (2, '2sigma', 0.954499736103642),
                                    (3, '3sigma', 0.997300203936740),
                                    (4, '4sigma', 0.999936657516334),
                                    (5, '5sigma', 0.999999426696856)])
def test_quantile_plot_interval_str(sigmas):
    qi1, qi2 = quantile_plot_interval(q=sigmas[0])
    qs1, qs2 = quantile_plot_interval(q=sigmas[1])
    assert qi1 == pytest.approx(0.5 - sigmas[2] / 2)
    assert qi2 == pytest.approx(0.5 + sigmas[2] / 2)
    assert qs1 == pytest.approx(0.5 - sigmas[2] / 2)
    assert qs2 == pytest.approx(0.5 + sigmas[2] / 2)


@pytest.mark.parametrize('floats', [0, 0.1, 0.9])
def test_quantile_plot_interval_float(floats):
    q1, q2 = quantile_plot_interval(q=floats)
    assert q1 == min(floats, 1 - floats)
    assert q2 == max(floats, 1 - floats)


@pytest.mark.parametrize('q1, q2', [(0, 1), (0.1, 0.9), (0, 0.9), (0.1, 1)])
def test_quantile_plot_interval_tuple(q1, q2):
    _q1, _q2 = quantile_plot_interval(q=(q1, q2))
    assert _q1 == q1
    assert _q2 == q2


@pytest.mark.parametrize('color', ['C0', 'k', 'gold', '#00FFFF',
                                   (1.0, 1.0, 0.0, 1.0)])
def test_basic_cmap(color):
    cmap = basic_cmap(color)

    # Check that the basic cmap is reversible
    cmap.reversed()
